use actix_web::{web, HttpResponse, Responder};
use serde::Deserialize;
use sqlx::PgPool;

use crate::{
    endpoints::assignment::PlanEventPersonalTablePartialTemplate,
    utils::{
        event_planning_template::{
            generate_availability_assignment_list, generate_status_whether_staff_is_required,
        },
        ApplicationError, TemplateResponse,
    },
};

use brass_db::{
    models::{Assignment, AssignmentChangeset, AssignmentContext, Event, Function, Role, User},
    validation::AsyncValidate,
};

#[derive(Deserialize)]
pub struct AssignmentQuery {
    availability: i32,
    function: u8,
    event: i32,
}

#[actix_web::post("/assignments/new")]
pub async fn post(
    user: web::ReqData<User>,
    pool: web::Data<PgPool>,
    query: web::Query<AssignmentQuery>,
) -> Result<impl Responder, ApplicationError> {
    if user.role != Role::Admin && user.role != Role::AreaManager {
        return Err(ApplicationError::Unauthorized);
    }

    let Some(event) = Event::read_by_id_including_location(pool.get_ref(), query.event).await?
    else {
        return Ok(HttpResponse::NotFound().finish());
    };

    let function = Function::try_from(query.function)?;

    let changeset = AssignmentChangeset {
        function,
        time: (event.start, event.end),
    };

    let context = AssignmentContext {
        user: &user.into_inner(),
        event_id: event.id,
        availability_id: query.availability,
        pool: pool.get_ref(),
    };

    if let Err(e) = changeset.validate_with_context(&context).await {
        return Ok(HttpResponse::UnprocessableEntity().body(e.to_string()));
    };

    Assignment::create(pool.get_ref(), event.id, query.availability, changeset).await?;

    let availabilities = generate_availability_assignment_list(pool.get_ref(), &event).await?;

    let (
        further_posten_required,
        further_fuehrungsassistent_required,
        further_wachhabender_required,
    ) = generate_status_whether_staff_is_required(pool.get_ref(), &event).await?;

    let template = PlanEventPersonalTablePartialTemplate {
        event,
        availabilities,
        further_posten_required,
        further_fuehrungsassistent_required,
        further_wachhabender_required,
    };

    Ok(template.to_response()?)
}

#[cfg(test)]
mod tests {
    use actix_http::StatusCode;
    use brass_db::models::{
        Area, Assignment, AssignmentChangeset, Availability, AvailabilityChangeset, Event,
        EventChangeset, Function, Location, Role, User, UserChangeset,
    };
    use brass_macros::db_test;
    use chrono::NaiveDateTime;
    use fake::{Fake, Faker};
    use sqlx::PgPool;

    use crate::utils::test_helper::{
        assert_snapshot, test_post, DbTestContext, NaiveDateTimeExt, RequestConfig,
        ServiceResponseExt,
    };

    async fn arrange(pool: &PgPool) {
        Location::create(pool, &Faker.fake::<String>(), 1)
            .await
            .unwrap();

        let mut user_changeset: UserChangeset = Faker.fake();
        user_changeset.name = String::from("Max Mustermann");

        User::create(pool, &user_changeset).await.unwrap();
    }

    async fn arrange_event(pool: &PgPool, start: NaiveDateTime, end: NaiveDateTime, location: i32) {
        let mut changeset: EventChangeset = EventChangeset::create_for_test(start, end);
        changeset.location_id = location;

        Event::create(pool, changeset).await.unwrap();
    }

    async fn arrange_availability(pool: &PgPool, start: NaiveDateTime, end: NaiveDateTime) {
        Availability::create(
            pool,
            1,
            AvailabilityChangeset {
                time: (start, end),
                comment: None,
            },
        )
        .await
        .unwrap();
    }

    #[db_test]
    fn response_produces_updated_template(context: &DbTestContext) {
        let app = context.app().await;

        let start = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 10, 0, 0).unwrap();
        let end = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 20, 0, 0).unwrap();

        arrange(&context.db_pool).await;
        arrange_event(&context.db_pool, start, end, 1).await;
        arrange_availability(&context.db_pool, start, end).await;

        let config = RequestConfig::new("/assignments/new?availability=1&function=1&event=1")
            .with_role(Role::Admin);

        let response = test_post::<_, _, String>(&context.db_pool, app, &config, None).await;
        let (status, body) = response.into_status_and_body().await;

        assert_eq!(StatusCode::OK, status, "{body}");

        assert_snapshot!(body);
    }

    #[db_test]
    fn fails_when_availability_does_not_exist(context: &DbTestContext) {
        let app = context.app().await;

        let start = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 10, 0, 0).unwrap();
        let end = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 20, 0, 0).unwrap();
        arrange(&context.db_pool).await;
        arrange_event(&context.db_pool, start, end, 1).await;

        let config = RequestConfig::new("/assignments/new?availability=1&function=1&event=1")
            .with_role(Role::Admin);

        let response = test_post::<_, _, String>(&context.db_pool, app, &config, None).await;

        assert_eq!(StatusCode::UNPROCESSABLE_ENTITY, response.status());
    }

    #[db_test]
    fn fails_when_event_does_not_exist(context: &DbTestContext) {
        let app = context.app().await;

        let start = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 10, 0, 0).unwrap();
        let end = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 20, 0, 0).unwrap();
        arrange(&context.db_pool).await;
        arrange_availability(&context.db_pool, start, end).await;

        let config = RequestConfig::new("/assignments/new?availability=1&function=1&event=1")
            .with_role(Role::Admin);

        let response = test_post::<_, _, String>(&context.db_pool, app, &config, None).await;

        assert_eq!(StatusCode::NOT_FOUND, response.status());
    }

    #[db_test]
    fn fails_when_area_manager_is_different_area_from_event(context: &DbTestContext) {
        let app = context.app().await;

        let start = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 10, 0, 0).unwrap();
        let end = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 20, 0, 0).unwrap();

        arrange(&context.db_pool).await;
        arrange_event(&context.db_pool, start, end, 1).await;
        arrange_availability(&context.db_pool, start, end).await;
        Area::create(&context.db_pool, &Faker.fake::<String>())
            .await
            .unwrap();

        let config = RequestConfig::new("/assignments/new?availability=1&function=1&event=1")
            .with_role(Role::AreaManager)
            .with_user_area(2);

        let response = test_post::<_, _, String>(&context.db_pool, app, &config, None).await;

        assert_eq!(StatusCode::UNPROCESSABLE_ENTITY, response.status());
    }

    #[db_test]
    fn fails_when_availability_user_not_in_event_area(context: &DbTestContext) {
        let app = context.app().await;

        let start = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 10, 0, 0).unwrap();
        let end = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 20, 0, 0).unwrap();

        arrange(&context.db_pool).await;
        Area::create(&context.db_pool, &Faker.fake::<String>())
            .await
            .unwrap();
        Location::create(&context.db_pool, &Faker.fake::<String>(), 2)
            .await
            .unwrap();
        arrange_event(&context.db_pool, start, end, 2).await;
        arrange_availability(&context.db_pool, start, end).await;

        let config = RequestConfig::new("/assignments/new?availability=1&function=1&event=1")
            .with_role(Role::Admin);

        let response = test_post::<_, _, String>(&context.db_pool, app, &config, None).await;

        assert_eq!(StatusCode::UNPROCESSABLE_ENTITY, response.status());
    }

    #[db_test]
    fn fails_assignment_time_doesnt_fit_into_availability_time(context: &DbTestContext) {
        let app = context.app().await;

        let start = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 10, 0, 0).unwrap();
        let end = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 20, 0, 0).unwrap();
        let availability_start = NaiveDateTime::from_ymd_and_hms(2025, 01, 12, 10, 0, 0).unwrap();
        let availability_end = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 18, 0, 0).unwrap();

        arrange(&context.db_pool).await;
        arrange_event(&context.db_pool, start, end, 1).await;
        arrange_availability(&context.db_pool, availability_start, availability_end).await;

        let config = RequestConfig::new("/assignments/new?availability=1&function=1&event=1")
            .with_role(Role::Admin);

        let response = test_post::<_, _, String>(&context.db_pool, app, &config, None).await;

        assert_eq!(StatusCode::UNPROCESSABLE_ENTITY, response.status());
    }

    // fn fails_when_end_time_lies_before_start_time(context: &DbTestContext) -> not possible as event has the same constraint

    #[db_test]
    fn fails_when_availability_time_already_assigned(context: &DbTestContext) {
        let app = context.app().await;

        let start = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 10, 0, 0).unwrap();
        let end = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 20, 0, 0).unwrap();

        arrange(&context.db_pool).await;
        arrange_event(&context.db_pool, start, end, 1).await;
        arrange_event(&context.db_pool, start, end, 1).await;
        arrange_availability(&context.db_pool, start, end).await;

        let assignment_changeset = AssignmentChangeset {
            function: Function::Posten,
            time: (start, end),
        };

        Assignment::create(&context.db_pool, 1, 1, assignment_changeset)
            .await
            .unwrap();

        let config = RequestConfig::new("/assignments/new?availability=1&function=1&event=2")
            .with_role(Role::Admin);

        let response = test_post::<_, _, String>(&context.db_pool, app, &config, None).await;

        assert_eq!(StatusCode::UNPROCESSABLE_ENTITY, response.status());
    }

    #[db_test]
    fn fails_when_availability_user_does_not_have_function(context: &DbTestContext) {
        let app = context.app().await;

        let start = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 10, 0, 0).unwrap();
        let end = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 20, 0, 0).unwrap();

        arrange(&context.db_pool).await;
        arrange_event(&context.db_pool, start, end, 1).await;
        arrange_availability(&context.db_pool, start, end).await;

        let config = RequestConfig::new("/assignments/new?availability=1&function=5&event=1")
            .with_role(Role::Admin);

        let response = test_post::<_, _, String>(&context.db_pool, app, &config, None).await;

        assert_eq!(StatusCode::UNPROCESSABLE_ENTITY, response.status());
    }

    #[db_test]
    fn fails_when_event_already_has_enough_assignments_for_function(context: &DbTestContext) {
        let app = context.app().await;

        let start = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 10, 0, 0).unwrap();
        let end = NaiveDateTime::from_ymd_and_hms(2025, 01, 10, 20, 0, 0).unwrap();

        Location::create(&context.db_pool, &Faker.fake::<String>(), 1)
            .await
            .unwrap();

        let mut user_changeset: UserChangeset = Faker.fake();
        user_changeset.functions.push(Function::Fuehrungsassistent);

        User::create(&context.db_pool, &user_changeset)
            .await
            .unwrap();
        User::create(&context.db_pool, &user_changeset)
            .await
            .unwrap();

        arrange_event(&context.db_pool, start, end, 1).await;

        Availability::create(
            &context.db_pool,
            1,
            AvailabilityChangeset {
                time: (start, end),
                comment: None,
            },
        )
        .await
        .unwrap();

        Availability::create(
            &context.db_pool,
            2,
            AvailabilityChangeset {
                time: (start, end),
                comment: None,
            },
        )
        .await
        .unwrap();

        let assignment_changeset = AssignmentChangeset {
            function: Function::Fuehrungsassistent,
            time: (start, end),
        };

        Assignment::create(&context.db_pool, 1, 1, assignment_changeset)
            .await
            .unwrap();

        let config = RequestConfig::new("/assignments/new?availability=2&function=5&event=1")
            .with_role(Role::Admin);

        let response = test_post::<_, _, String>(&context.db_pool, app, &config, None).await;

        assert_eq!(StatusCode::UNPROCESSABLE_ENTITY, response.status());
    }
}
